
## cluster analysis
##
## run at the R prompt source(file = "cluster.R", echo = TRUE)
##
## (C) ceeboo 2008, 2009

## Compute cosine distances between all rows of x and y,
## The low-level implementation maps the similarity of two
## all-zero (and all-NA) vectors to one. Notice that pairs of
## components with missing values are omitted in the nominator,
## and if there are no components left a missing value is
## returned. However, in the denominator, missing values are
## omitted independently from x and y.
## 
require(proxy)
cosine.xy <-
function(x, y = NULL, pairwise = FALSE)
    1 - .Call("R_cosine", x, y, pairwise)

## Compute the means of the equivalence classes specified
## in k. Missing values are omitted possibly biasing the
## means to higher or lower values.
mean.k <-
function(x, k)
    apply(x, 2, tapply, k, function(x) {
        x <- mean(x, na.rm = TRUE)
        if(is.nan(x)) NA else x
    })

## Partition data into equivalence classes such that the
## total sum of the distances (from the winning centroid)
## is minimized. init can be a matrix of initial centroids,
## a scalar value specifiying their number, or a vector
## classifiying the rows of data. For further references
## see help(topic = "kmeans") and R package "flexclust".
kcentroids <-
function(data, init, DFUN, CFUN = mean.k, niter = 100, nstart = 1) {
    if (!is.matrix(init))
        init <- if(length(init) == nrow(data))
                    CFUN(data, init)
                else
                    data[sample(seq_len(nrow(data)), init),,
                        drop = FALSE]
    l <- NULL
    while (niter > 0) {    
        k <- factor(apply(DFUN(data, init), 1, function(x) {
            x <- which.min(x)
            if(length(x)) x else NA
        }), levels = seq_len(nrow(init)))
        if (identical(k, l))
            break
        l <- k
        init <- CFUN(data, k)
        niter <- niter - 1
    }
    if (niter == 0)
        warning("no convergence")
    r <- DFUN(data, init[k,, drop = FALSE], TRUE)
    f <- mean(r, na.rm = TRUE)
    if (is.nan(f))
        f <- NA
    r <- list(centers = init, cluster = k, distance = r, size = table(k),
        fit = f)
    while (nstart > 1) {
        t <- kcentroids(data, nrow(init), DFUN, CFUN)
        if (t$fit < r$fit)
            r <- t
        nstart <- nstart - 1
    }
    r
}

## Compute the table of counts of first and second
## winners. Omit data points whose distance exceeds
## the threshold c. 
nbhtable <- 
function(data, centers, DFUN, k, c = NULL) {
    if (missing(k)) {
        k <- rownames(centers)
        if (is.null(k))
            k <- seq(nrow(centers))
    }
    r <- DFUN(data, centers)
    if (!is.null(c))
        is.na(r) <- r > c
    r <- apply(r, 1, function(x) {
        i <- order(x, na.last = TRUE, decreasing = FALSE)
        is.na(i) <- is.na(x[i])
        k[i[1:2]]
    })
    table(First = factor(r[1, ], levels = k),
          Second = factor(r[2, ], levels = k))
}

## Scale the rows of a matrix to sum to unity.
rowScale <-
function(x, na.rm = FALSE)
{
    s <- rowSums(x, na.rm = na.rm)
    x / (s + (s == 0))
}

## Compute the Simple Structure Index, optionally
## weighted by the number of contributing centroids,
## and/or scaled to the attribute ranges of data.
## Missing values are omitted.
ssi <- function(centers, weight = FALSE, data) {
    r <- apply(centers, 2, function(x) {
        l <- which.min(x)
        l <- if(length(l)) x[l] else NA
        u <- which.max(x)
        u <- if(length(u)) x[u] else NA
        c(l, u)
    })
    s <- r[2, ] - r[1, ]
    s <- sum(s, na.rm = TRUE) / length(s)
    if (weight) {
        x <- apply(t(centers) == r[1, ] | t(centers) == r[2, ],
            2, any, na.rm = TRUE)
        s <- s * sum(x, na.rm = TRUE) / length(x)
    }
    if (!missing(data))
        s <- s / ssi(data)
    s
}

fmt.y <- function(x) {
    if (length(x) > 1)
        x <- tapply(x, cumsum(c(TRUE, abs(diff(x)) > 1)), function(x)
            if (length(x) > 1)
                paste(x[1], abs(x[length(x)]), sep = ":")
            else
                as.character(x))
    paste(x, collapse = ",")
}

###
load(file = "city_ALL_95-07.rda")

assign("data", city_ALL)

## Reduce time frame
data <- data[,, as.character(1995:2006)]

## reduce set of cities
#k <- match(c("BER","MAD","OSL","STO","HEL"), dimnames(data)[[1]])
#data <- data[-k,,]

## Stack market profiles of cities
ds <- matrix(as.double(aperm(data, c(2,1,3))), ncol = dim(data)[2],
    dimnames = dimnames(data)[c(NA, 2)], byrow = TRUE)
## Row indexes
ds.id <- list(Year = rep(as.integer(dimnames(data)[["Year"]]),
    each = dim(data)[1]), City = rep(dimnames(data)[["City"]], dim(data)[3]))

## Map profiles with too many missing values to the
## information-less profile thereby eliminating them
## from further analysis.
#k <- apply(is.na(ds), 1, sum)
#cumsum(rev(table(k)))
#ds[k > 6,] <- NA_integer_

## Find an initial solution by hierarchical clustering.
## Since hclust does not work with missing values we map
## to zero, thereby clustering the "incomparables".
d <- cosine.xy(ds)
d[is.na(d)] <- 0
hc <- hclust(d, method = "average")

## inspect visually
plot(hc, labels = FALSE, hang = -1, main = "Guest Mix Dendrogram")
rect.hclust(hc, k = 17)

## check for necessary level of detail
nc <- lapply(7:27, function(k) 
    kc <- kcentroids(ds, cutree(hc, k = k), cosine.xy)
)
t <- data.frame(fit = sapply(nc, "[[", "fit"),
    SSI = sapply(nc, function(x) ssi(x$centers, TRUE, ds)))
rownames(t) <- 7:27

op <- par(mfrow = c(2, 1))
plot(7:27, rev(hc$height)[7:27], type = "b", ylab = "Height",
    xlab = "Number of Clusters", main = "hclust")
abline(v = c(10, 15, 20, 25), lty = 2, lwd = 0.1)
plot(7:27, t$SSI, type = "b", ylab = "SSI",
     main = "k-centroids", xlab="Number of clusters")
abline(v = c(10, 15, 20, 25), lty = 2, lwd = 0.1)
par(new=TRUE)
plot(7:27, t$fit, type = "b", ylab = "", xlab = "", pch = 24,
     sub = "", main = "", axes = FALSE, lty = 2, col = 2)
axis(4, col = 2)
par(op)

## Cut the tree at k clusters. 
k <- cutree(hc, k = 17)

t <- tapply(seq_len(length(k)), k, function(k) summary(d[[k]]))
t <- sapply(t[!sapply(t, is.null)], eval)
t <- rbind(t * 100, Size = table(k))
names(dimnames(t)) <- c("Distance", "Cluster")
print(t(t), digits = 1)

## Final solution
kc <- kcentroids(ds, k, cosine.xy)

## Average distance
kc$fit

## Simple structure index.
ssi(kc$centers, TRUE, ds)

## Check homogeneity of clusters
t <- tapply(kc$distance, kc$cluster, summary)
t <- sapply(t[!sapply(t, is.null)], eval)
t <- rbind(t * 100, Size = kc$size)
names(dimnames(t)) <- c("Distance", "Cluster")
print(t(t), digits = 1)

## Distances between points of a cluster
t <- tapply(seq_len(length(kc$cluster)), kc$cluster,
    function(i) summary(c(cosine.xy(ds[i,,drop=FALSE]))))
t <- sapply(t, eval)
t <- rbind(t * 100, Size = kc$size)
names(dimnames(t)) <- c("Distance", "Cluster")
print(t(t), digits = 1)

## Check the proportion (bias) of missing values
t <- apply(ds, 2, tapply, kc$cluster, function(x) mean(is.na(x)))
names(dimnames(t)) <- c("Cluster", "Market")
print(t(t) * 100, digits = 1)

## Number of data points with number of missing
## values contained in the clusters
t <- table(apply(ds, 1, function(x) sum(is.na(x))),
    factor(kc$cluster, exclude = NULL))
t <- cbind(t, Total = rowSums(t))
t <- rbind(t, Total = colSums(t))
names(dimnames(t)) <- c("NA's", "Cluster")
t

## Centroids rescaled to market shares
t <- rowScale(kc$centers, TRUE)
names(dimnames(t)) <- c("Cluster", "Market")
print(t(t) * 100, digits = 1)

## Compute entropy (diversification).
k <- apply(t, 1, function(x) -sum(x * log(x), na.rm = TRUE))
round(k, 2)


## Distribution of cities across clusters
t <- table(ds.id$City, factor(kc$cluster, exclude = NULL))
n <- apply(t, 2, function(x) sum(x > 0))
t <- rbind(t, Total = colSums(t))
t <- rbind(t, City = n)
names(dimnames(t)) <- c("City", "Cluster")
t

## Dominant markets (see below)
d <- as.list(apply(kc$centers, 1, which.max))
s <- as.list(apply(rowScale(kc$centers, TRUE), 1, max, na.rm = TRUE))
c <- as.list(rep(1, length(levels(kc$cluster))))
m <- as.list(apply(kc$centers, 1, function(x) sum(is.na(x))))
z <- as.list(apply(rowScale(kc$centers, TRUE), 1,
     function(x) -sum(x * log(x), na.rm = TRUE)))
	

## Check for differences in competing cities.
t <- table(ds.id$City, kc$cluster)
n <- apply(t, 2, function(x) sum(x > 0))
for (j in which(kc$size & n > 1)) {
    i <- kc$cluster == j
    t <- mean.k(ds[i, ], ds.id$City[i])
    m[[j]] <- apply(t, 1, function(x) sum(is.na(x)))
    cat(paste(formatC(j, width = 2), ":", sep = ""),
        formatC(m[[j]], width = 2), "\n")
    ## Dominant markets
    d[[j]] <- k <- apply(t, 1, which.max)
    s[[j]] <- rowScale(t, na.rm = TRUE)[cbind(seq_len(length(k)) ,k)]
    k <- apply(t, 1, sum, na.rm = FALSE)
    c[[j]] <- k / sum(k, na.rm = TRUE)
    z[[j]] <- apply(rowScale(t, TRUE), 1,
	 function(x) -sum(x * log(x), na.rm = TRUE))

}

## Summarize city / year distribution across clusters
t <- mapply(tapply, tapply(ds.id$Year, kc$cluster, list),
        tapply(ds.id$City, kc$cluster, list),
    MoreArgs = list(FUN = eval))
t <- lapply(t, lapply, function(x, k)
        if (length(x) <= length(k) / 2) x else -k[-match(x, k)],
    as.integer(dimnames(data)[["Year"]]))
n <- sapply(t, length)

t <- data.frame(
    Cluster = rep(names(t), n),
    City    = unlist(sapply(t, names)),
    Year    = unlist(lapply(t, sapply, fmt.y)),
    Volume  = round(unlist(c) * 100),
    Market  = dimnames(data)[["Market"]][unlist(d)],
    Share   = round(unlist(s) * 100, 1),
    NAs     = unlist(m),
    DI      = round(unlist(z), 2),
    row.names = seq(sum(n)))
t

## informationless cluster
k <- is.na(kc$cluster)
k <- tapply(ds.id$Year[k], ds.id$City[k], fmt.y)
data.frame(
    Cluster = rep(NA_integer_, length(k)),
    City    = names(k),
    Year    = structure(k),
    Volume  = rep(NA_real_, length(k)),
    Market  = rep(NA_character_, length(k)),
    Share   = rep(NA_real_, length(k)),
    NAs     = rep(ncol(ds), length(k)),
    DI      = rep(NA_real_, length(k)),
    row.names = seq_len(length(k)) + nrow(t)
)


## compute cluster label of nearest neighbors 
k <- apply(as.matrix(cosine.xy(ds), diag = NA), 1,
     function(d) {
	k <- which.min(d)
	if (!length(k)) NA else kc$cluster[k]
     })
t <- table(kc$cluster, factor(k, levels = levels(kc$cluster)))
t

## Check topology of centroids
dc <- cosine.xy(kc$centers)
print(dc * 100, digits = 1)

## Check neighborhood relations
nt <- nbhtable(ds, kc$centers, cosine.xy, c = 0.5)
print(nt)


k <- table(kc$cluster)
## relative to cluster size
round(nt / as.vector(k) * 100)
## relative to joint cluster size
t <- (nt + t(nt)) / outer(k, k, "+")
t <- as.dist(round(t * 100))
t

## relative combined market volume
k <- apply(kc$centers, 1, sum, na.rm = TRUE)
k <- round(k / max(k) * 100)
k

## Visualize market(s)
if (require("MASS") && all(kc$size > 0)) {
    s <- sammon(dc)
#s$points[,1] <- -s$points[,1]
    plot(s$points, type = "n", xlab = "", ylab = "",
        main = "City Market Map")
    abline(v = 0, h = 0, lty = 3)
#    t <- as.dist(t(nt) + nt)
    is.na(t) <- t == 0
    t <- cut(c(t), breaks = 4)
    print(levels(t))
    l <- !is.na(t)
    i <- row.dist(dc)[l]
    j <- col.dist(dc)[l]
    segments(s$points[i, 1], s$points[i, 2],
             s$points[j, 1], s$points[j, 2],
        lwd = t[l], lty = (c(dc)[l] > 0.5) + 1)

    m <- apply(kc$centers, 1, which.max)
    l <- sort(unique(m))
    col <- NULL
#    col[l] <- rainbow(dim(data)[2])[l]
    col[l] <- rainbow(length(l))
    points(s$points, cex = kc$size / dim(data)[3] * 3,
        pch = 19, col = col[m])

    c <- tapply(ds.id$City, kc$cluster, function(x) {
	    x <- unique(x)
	    if (length(x) > 1)
		x <- paste(substr(x, 1, 1), collapse = "")
	    x
    })
    text(s$points, labels = c, cex = 0.8)

    text(s$points, labels = k, adj = c(2,-1))

    text((s$points[i, ] + s$points[j, ]) / 2,
	labels = paste(nt[cbind(i,j)],nt[cbind(j,i)], sep = ":"),
	cex = 0.6, adj = c(1.2,-0.5))

    legend("topleft", legend = dimnames(data)[["Market"]][l], cex = 0.7,
        fill = col[l], bty = "n")
}

###
