
## (C) ceeboo 2007, 2008, 2009

## indscal 3-mode 2-way 
##
## see Borg, Ingwer and Groenen, Patrick J. F. 2005. Modern multidimensional
##     scaling theory and applications. Springer, New York, pp. 385-386.  
##
## FIXME convergence is slow

.scale <-
function(x, scale = FALSE)
{
    n <- dim(x)[1]
    H <- diag(n) - outer(rep(1,n), rep(1,n)) / n

    b <- apply(x, 3, function(x) H %*% -(x^2/2) %*% H / 2)
    b <- array(b, dim = dim(x))
    if (scale) {
        b <- apply(b, 3, function(x) x / sqrt(sum(x^2) / n / (n - 1)))
        b <- array(b, dim = dim(x))
    }
    b
}

indscal <-
function(d, dim = 2, scale = FALSE, crit = 1e-6, maxIter = 1e5, verbose = FALSE)
{
    ## center
    b <- .scale(d, scale = scale)

    ## average over subjects
    B <- apply(b, c(1,2), mean)

    ## decompose
    s <- eigen(B, symmetric = TRUE)

    x <- s$vectors[,seq(dim)]
   
    if (verbose)
        cat("eigenvalues: ", s$values[seq(dim)], "\n\n",
            "  #   ITER        NORM                FIT\n")

    iter <- best <- 0
    while (iter < maxIter) {
        ## fit weights
        B <- NULL
        for (k in seq(dim(b)[3]))
            B <- cbind(B, c(b[,,k]))

        V <- NULL
        for (j in seq(dim(x)[2]))
            V <- cbind(V, c(outer(x[,j], x[,j])))

        # solve B = V W
        q <- qr(V)
        w <- qr.coef(q, B)

        if (any(w < 0))
            stop("negative dimension weight(s)")

        # rearrange
        B <- NULL
        for (k in seq(dim(b)[3]))
            B <- rbind(B, b[,,k])

        xl <- x

        step <- last <- 0
        while (step < maxIter * 4) {

            ## fit right coordinates
            L <- NULL
            for (j in seq(dim(w)[2]))
                L <- rbind(L, xl %*% diag(w[,j]))

            # solve B = L X'
            q  <- qr(L)
            xr <- qr.coef(q, B)
            xr <- t(xr)

            ## fit left coordinates
            R <- NULL
            for (j in seq(dim(w)[2]))
                R <- rbind(R, xr %*% diag(w[,j]))

            # solve B = R X'
            q  <- qr(R)
            xl <- qr.coef(q, B)
            xl <- t(xl)

            g  <- sum((xl-xr)^2)

            if (step && last-g < crit)
                break

            last <- g
            step <- step + 1
        }
        # stabilize
        x  <- (xl + xr) / 2

        f <- 0
        for (j in seq(dim(w)[2]))
            f <- f + sum((b[,,j] - x %*% diag(w[,j]) %*% t(x))^2)

        if (verbose)
            cat(formatC(iter, width = 4),
                if (best < f) "~" else " ",
                formatC(step, width = 4),
                if (last < g) "~" else " ",
                formatC(g,    width =  9, digits = 6, "f"),
                formatC(f,    width = 18, digits = 6, "f"),
                        "\n")

        if (iter && best-f < crit)
            break

        best <- f
        iter <- iter + 1
    }

    w <- t(w)

    dimnames(w) <- c(dimnames(d)[3], list(dim = seq(dim)))
    dimnames(x) <- c(dimnames(d)[2], list(dim = seq(dim)))

    ## package result
    r <- list(call    = match.call(),
              fit     = f,
              iter    = iter,
              weights = w,
              points  = x)
    class(r) <- "indscal"

    r
}

##
load(file = "city_ALL_95-07.rda")
assign("data", city_ALL)


## reduce time frame
data <- data[,,as.character(1995:2006)]
dd <- as.double(data)
attributes(dd) <- attributes(data)

## determine the number of profiles we lose
## for different thresholds of the number of
## missing values per profile
cumsum(rev(table(apply(dd, 3, apply, 1, function(x) sum(is.na(x))))))

## Map profiles with too many missing values to the
## information-less profile thereby producing NA
## distance values, which we later predict from the
## usable data.
dd <- apply(dd, 3, apply, 1, function(x) {
    if (sum(is.na(x)) > 6)
	 x[] <- NA_real_
    x
})
dd <- aperm(array(dd, dim = dim(data)[c(2,1,3)],
     dimnames = dimnames(data)[c(2,1,3)]), c(2, 1, 3))
## check
table(dd == data, exclude = NULL)

## compute cosine similarities (see the
## explanation in cluster.R)
require(proxy)
cosine.xy <-
function(x, y = NULL, pairwise = FALSE)
    1 - .Call("R_cosine", x, y, pairwise)

## compute distances
dd <- array(apply(dd, 3, function(x) as.matrix(cosine.xy(x), diag = 0)),
    dim = dim(dd)[c(1,1,3)], dimnames = dimnames(dd)[c(1,1,3)])
table(is.na(dd))

## check bias due to missing values
print(apply(dd, 3, function(x) sum(is.na(x))))
print(apply(dd, 1, function(x) sum(is.na(x))))

## predict missing distance using the
## average over time 
for (n in dimnames(data)[["City"]]) {
    m <- apply(dd[n,,], 1, median, na.rm = TRUE)
    k <- is.na(dd[n,,])
    dd[n,,][k] <- rep(m, dim(data)[3])[k]
}

## fit INDSCAL model
s <- indscal(dd, dim = 3, scale = TRUE)
s$fit

## visualize weights
matplot(rownames(s$weights), s$weights, type = "l",
        ylab = "Weight", xlab = "Year",
        lty = seq(ncol(s$weights)), col = seq(ncol(s$weights)),
        main = "Indscal Dimension Weights")
grid()
legend("bottomright", legend = seq(ncol(s$weights)),
    col = seq(ncol(s$weights)), lty = seq(ncol(s$weights)), bty = "n")

## mean weights
m <- colMeans(s$weights)
print(m, digits = 2)
## all weights
print(s$weights, digits = 1)


## interactive 3d plot
if (require(rgl)) {
    ## extend the the x and y plotting ranges
    ## have grid lines at all boundaries.
    p <- s
    p$points[,2] <- -p$points[,2]
    plot3d(p$points, type = "h", box = FALSE, 
	 xlim = c(-0.65,0.25), ylim = c(-0.65, 0.45), axes = F,
	 xlab = "", ylab = "", zlab = "",
	 sub = "Indscal City Space")
    ## indicate center in x-y plane
    grid3d(side = "z", at = list(z = 0), n = 1, lty = 3, col = "red")
    grid3d(side = "z", at = list(z = 0))
    points3d(cbind(p$points[,1:2],0))
    text3d(p$points, text = rownames(p$points), cex = 0.6)
    #view3d(theta = , phi = )
}

## compute correlations of map positions
## with markets

## no missing values in this year
z <- data[,,"2003"]
z <- apply(z, 1, function(x) x /sum(x))
r <- cor(t(z), s$points, use = "pairwise.complete.obs")
print(r, digits = 2)

## use all years (and compare to above)
m <- NULL
for (j in seq(ncol(s$weights)))
    m <- cbind(m, as.vector(outer(s$points[,j], s$weights[,j])))
colnames(m) <- colnames(s$points)

z <- apply(data, 3, apply, 1, function(x)
    if (sum(is.na(x)) > 6)
	rep(NA_real_, length(x))
    else
	x / sum(x, na.rm = TRUE)
)
z <- matrix(z, ncol = dim(data)[2], byrow = TRUE)
colnames(z) <- dimnames(data)[[2]]

r <- cor(z, m, use = "pairwise.complete.obs")
names(dimnames(r)) <- c("Market", "Dimension")
round(r, digits = 2)

## determine insignificant correlations
t <- apply(m, 2, function(y)
     apply(z, 2, function(x) cor.test(x, y)$p.value))
t >= 0.05

## compute development of Euclidean distances
## between cities over time
t <- apply(s$weights, 1, function(w) {
    d <- dist(s$points * rep(w, each = dim(s$points)[1]))
    c(d)
})

## average distances
d <- structure(rowMeans(t), Size = dim(s$points)[1],
    Labels = rownames(s$points), class = "dist")

op <- par(mfrow = c(3, 1))

k <- row.dist(d) == match("ZAG", rownames(s$points)) &
     col.dist(d) == match("LIS", rownames(s$points))

plot(colnames(t), t[k, ], type = "b", ylab = "Distance", xlab = "Year",
    main = "Lisbon / Zagreb")

k <- row.dist(d) == match("ZAG", rownames(s$points)) &
     col.dist(d) == match("HEL", rownames(s$points)) 

plot(colnames(t), t[k, ], type = "b", ylab = "Distance", xlab = "Year",
    main = "Helsinki / Zagreb")

k <- row.dist(d) == match("LIS", rownames(s$points)) &
     col.dist(d) == match("HEL", rownames(s$points)) 

plot(colnames(t), t[k, ], type = "b", ylab = "Distance", xlab = "Year",
    main = "Helsinki / Lisbon")
par(op)


## dump distance data for use with other
## statistical software
x <- matrix(dd, ncol = dim(dd)[1], byrow = TRUE)
colnames(x) <- dimnames(dd)[[1]]
x <- cbind(Year = rep(as.integer(dimnames(dd)[[3]]), each = dim(dd)[1]),
    as.data.frame(x))
write.csv(x, file = "city_93-06_dst.csv", row.names = FALSE)

###
